import click
import datetime
import sys
import tabulate
import os

from migrator.config import Config
from migrator.executor import Executor

from migrator.fixtures import FixtureLoader


def choices_prompt(message, choices, default='0'):
    return click.prompt(
        '{}\n{}'.format(
            '\n'.join(['{} - {}'.format(k, v) for k, v in sorted(choices.items(), key=lambda x: x[0])]),
            message
        ),
        default=default, type=click.Choice(sorted(choices.keys())), show_default=True
    )


def halt(messgae):
    click.echo(messgae)
    sys.exit(1)


def get_one_revision(migrator, rev):
    revisions = migrator.get_migrations_by_hash(rev)
    if not revisions:
        halt(f'Revision {rev} not found.')
    if len(revisions) > 1:
        halt(f'Revision {rev} has too much matches ({len(revisions)}).')
    return revisions[0]


@click.group()
@click.option('-c', '--config', type=click.Path(), default='migrator.cfg')
@click.pass_context
def cli(ctx, config):
    config = os.path.realpath(config)
    cfg = Config()
    ctx.obj = ctx.obj or {}
    if os.path.exists(config):
        cfg.load(config)
        ctx.obj['config_path'] = config
    else:
        cfg.make_default()
        ctx.obj['config_path'] = None
    ctx.obj.update({'cfg': cfg})


@cli.command()
@click.option('--type', 'db_type_arg', default=None)
@click.option('--name', 'db_name_arg', default=None)
@click.option('--user', 'db_user_arg', default=None)
@click.option('--password', 'db_password_arg', default=None)
@click.option('--host', 'db_host_arg', default=None)
@click.option('--port', 'db_port_arg', default=None)
@click.option('--dir', 'migrations_dir_arg', default=None)
@click.option('--sys_path', 'sys_path_arg', default=None)
@click.option('--models', 'models_path_arg', default=None)
@click.option('--excluded', 'models_excluded_arg', default=None)
@click.option('--storage', 'migrations_storage', default=None)
@click.option('--force', default=False, is_flag=True)
@click.pass_context
def create_config(
        ctx, db_type_arg, db_name_arg, db_user_arg, db_password_arg, db_host_arg, db_port_arg,
        migrations_dir_arg, sys_path_arg, models_path_arg, models_excluded_arg, migrations_storage, force
):
    config_path = ctx.obj['config_path'] or 'migrator.cfg'
    ask_edit = all(
        x is None for x in (
            db_type_arg, db_name_arg, db_user_arg, db_password_arg, db_host_arg, db_port_arg,
            migrations_dir_arg, sys_path_arg, models_path_arg, models_excluded_arg
        )
    )
    if os.path.exists(config_path):
        if not force and not click.confirm('Config already exists. Overwrite?', default=False):
            return
    db_types = {'0': 'postgres', '1': 'sqlite', '2': 'mysql'}

    if db_type_arg is None:
        result = choices_prompt('Database type', db_types)
        db_type = db_types[result]
    else:
        db_type = db_type_arg
    click.echo(db_type)
    db_name = (
        click.prompt('Database', default='peewee{}'.format('.db' if db_type == 'sqlite' else ''))
        if db_name_arg is None else db_name_arg
    )
    if db_type == 'sqlite':
        db_url = 'sqlite:///{}'.format(db_name)
    else:
        user = click.prompt('User', default='peewee') if db_user_arg is None else db_user_arg
        password = click.prompt('Password', hide_input=True) if db_password_arg is None else db_password_arg
        host = click.prompt('Host', default='127.0.0.1') if db_host_arg is None else db_host_arg
        port = (
            click.prompt('Port', default='5432' if db_type == 'postgres' else '3306')
            if db_port_arg is None else db_port_arg
        )
        db_url = '{}://{}:{}@{}:{}/{}'.format(db_type, user, password, host, port, db_name)

    default_path = os.path.abspath('.')
    migrations_dir = (
        click.prompt(
            'Migrations directory', default=os.path.join(default_path, 'migrations'), type=click.Path(
                exists=True, file_okay=False, writable=True, readable=True, resolve_path=True
            )
        ) if migrations_dir_arg is None else migrations_dir_arg
    )
    sys_path = (
        click.prompt(
            'PYTHON_PATH for project (Use : as delimiter)', default=default_path, type=click.Path(
                exists=True, file_okay=False, writable=False, readable=True
            )
        ) if sys_path_arg is None else sys_path_arg
    )

    models_path = (
        click.prompt('Models path (Comma separated)', default='app.models')
        if models_path_arg is None else models_path_arg
    )

    excluded_models = (
        click.prompt('Excluded models by name (Comma separated)', default='')
        if models_excluded_arg is None else models_excluded_arg
    )

    storage_types = {'fs': 'File system', 'db': 'Database'}

    if migrations_storage not in storage_types:
        migrations_storage = choices_prompt('Applied migrations storage type', storage_types, default='fs')

    # Add to new config
    cfg = Config()
    cfg.update({
        cfg.BASE_SECTION: {
            cfg.MIGRATOR_DB_URL: db_url,
            cfg.MIGRATOR_SYS_PATH: sys_path,
            cfg.MIGRATOR_MIGRATIONS_DIR: migrations_dir,
            cfg.MIGRATOR_MODELS_PATH: models_path,
            cfg.MIGRATOR_EXCLUDED_MODELS: excluded_models,
            cfg.MIGRATOR_MIGRATIONS_STORAGE: migrations_storage
        }
    })

    config = cfg.to_string_io().getvalue()

    if ask_edit and click.confirm('Edit config?', default=True):
        edited = click.edit(config)
        if edited is not None:  # None - if edit canceled
            config = edited
    with open(config_path, 'w', encoding='utf-8') as f:
        f.write(config)
    click.echo('Config successfully saved!')


@cli.command('make')
@click.option('--type', 'migration_type', default='from_db',
              type=click.Choice(['from_db', 'from_last', 'from_rev', 'empty', 'data']))
@click.option('--rev', default=None)
@click.option('--name', default=None)
@click.option('--models', default=None)
@click.option('--require', default=False, is_flag=True)
@click.pass_context
def make_migration(ctx, migration_type, rev, name, models, require):
    migrator = Executor(ctx.obj['cfg'])
    if rev is None and migration_type == 'rev':
        halt('--rev param required')
    migration_name = (click.prompt('Migration title') or None) if name is None else name
    revision = None

    if migration_type == 'empty':
        revision = migrator.make_empty_migration(migration_name=migration_name)
    elif migration_type == 'from_db':
        revision = migrator.migrate_from_db(migration_name=migration_name)
    elif migration_type == 'from_last':
        last = sorted(migrator.get_migrations(), key=lambda x: -x['time'])
        if not last:
            halt('There are no latest migration.')
        revision = migrator.migrate_from_migration(migration=last[0], migration_name=migration_name)
    elif migration_type == 'from_rev':
        revisions = migrator.get_migrations_by_hash(rev)
        if not revisions:
            click.echo(f'Revision {rev} not found.')
        if len(revisions) > 1:
            halt(f'Revision {rev} has too much matches ({len(revisions)}).')
        revision = migrator.migrate_from_migration(migration=revisions[0], migration_name=migration_name)
    elif migration_type == 'data':
        loader = FixtureLoader(ctx.obj['cfg'])
        revision = loader.make_data_migration(migration_name=migration_name, only_models=models)
    else:
        print(f'Unknown migration type {migration_type}')
    if require and revision is not None:
        migrator.make_required(revision)
        click.echo(f'Revision {revision} marked as required')


@cli.command('apply')
@click.option('--force', default=False, is_flag=True)
@click.option('--fake', default=False, is_flag=True)
@click.argument('rev')
@click.pass_context
def apply_migration(ctx, force, fake, rev):
    migrator = Executor(ctx.obj['cfg'])
    revision = get_one_revision(migrator, rev)
    if migrator.check_status(revision['hash']) == migrator.STATUS_APPLIED:
        if not force and not click.confirm('Migration already applied. Repeat?', default=False):
            halt('Abort')
    if not migrator.check_dependencies(revision):
        not_applied = [x for x in revision['dependencies'] if migrator.check_status(x) != migrator.STATUS_APPLIED]
        halt('Migration dependencies not applied: {}'.format(','.join(not_applied)))
    migrator.apply(revision, fake=fake)
    click.echo(f'Migration {revision["hash"]} applied successfully!')


@cli.command('revert')
@click.option('--force', default=False, is_flag=True)
@click.option('--fake', default=False, is_flag=True)
@click.argument('rev')
@click.pass_context
def revert_migration(ctx, force, fake, rev):
    migrator = Executor(ctx.obj['cfg'])
    revision = get_one_revision(migrator, rev)
    if migrator.check_status(revision['hash']) != migrator.STATUS_APPLIED:
        if not force and not click.confirm('Migration not applied. Revert?', default=False):
            halt('Abort')
    migrator.revert(revision, fake=fake)
    click.echo(f'Migration {revision["hash"]} reverted successfully!')


@cli.command('list')
@click.option('--sort', default='time', type=click.Choice(['name', 'time', 'hash', 'status']))
@click.option('--reverse', default=False, is_flag=True)
@click.pass_context
def migrations_list(ctx, sort, reverse):
    migrator = Executor(ctx.obj['cfg'])
    headers = ['Time', 'Title', 'Migration', 'Dependencies', 'Applied', 'Required']
    rows = []
    for migration in migrator.get_migrations():
        rows.append([
            datetime.datetime.fromtimestamp(migration['time']),
            migration['name'],
            migration['hash'],
            ', '.join(migration['dependencies']) if migration['dependencies'] else 'No',
            'Yes' if migration['status'] == migrator.STATUS_APPLIED else 'No',
            '{}, {}'.format('Yes', migration['required'] + 1) if migration['required'] is not None else 'No'
        ])
    sort_column = {'time': 0, 'name': 1, 'hash': 2, 'status': 4}[sort]
    rows = sorted(rows, key=lambda x: x[sort_column])
    if reverse:
        rows = reversed(rows)
    click.echo(tabulate.tabulate(rows, headers, tablefmt='psql'))


@cli.command('require')
@click.option('--after', default=None)
@click.argument('rev')
@click.pass_context
def mark_required(ctx, after, rev):
    migrator = Executor(ctx.obj['cfg'])
    revision = get_one_revision(migrator, rev)['hash']
    if after is not None:
        after = get_one_revision(migrator, after)['hash']
    migrator.make_required(revision, after=after)
    click.echo(f'Revision {revision} marked as required')


@cli.command('up')
@click.option('--fake', default=False, is_flag=True)
@click.pass_context
def up_required(ctx, fake):
    migrator = Executor(ctx.obj['cfg'])
    required = migrator.get_required()
    if not required:
        halt('There are no required migrations (Empty required.json)')

    to_apply = [rev for rev in required if migrator.check_status(rev) != migrator.STATUS_APPLIED]
    if not to_apply:
        halt('Project already up to date')
    for rev in to_apply:
        revision = get_one_revision(migrator, rev)
        if not migrator.check_dependencies(revision):
            not_applied = [x for x in revision['dependencies'] if migrator.check_status(x) != migrator.STATUS_APPLIED]
            halt('Migration dependencies not applied: {}'.format(','.join(not_applied)))
        migrator.apply(revision, fake=fake)
        click.echo(f'Migration {revision["hash"]} applied successfully!')


if __name__ == '__main__':
    cli()



import hashlib
import json
import os
import time

import peewee
from playhouse.shortcuts import model_to_dict, dict_to_model

from migrator.code_generator import CodeGenerator
from migrator.db_inspector import Inspector
from migrator.executor import Executor
from migrator.utils import SafeEncoder, pickle_hook


class FixtureLoader(object):
    # Pycharm stub
    _table = None
    _with_schema = None

    def __init__(self, config):
        self.config = config
        self.executor = Executor(config=config)
        self.db = self.config.get_db()

    def make_data_migration(self, migration_name, only_models=None):
        only_models = only_models.split(',') if only_models else []
        inspector = Inspector(excluded_models=self.config.get_excluded())

        data_to_save = {}
        for model_class in inspector.get_database_models(self.db):
            class_name = model_class.__name__
            if only_models and class_name not in only_models:
                continue
            data = [model_to_dict(obj) for obj in model_class.select()]
            if data:
                data_to_save[class_name] = data

        migration_time = int(time.time())
        migration_hash = hashlib.md5(str(migration_time).encode('utf-8')).hexdigest()
        fixture_path = os.path.join(self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR), 'fixtures')
        os.makedirs(fixture_path,  exist_ok=True)
        fixture_file_name = os.path.join(fixture_path, '{}.json'.format(migration_hash))
        with open(fixture_file_name, 'w', encoding='utf-8') as f:
            f.write(json.dumps(data_to_save, cls=SafeEncoder, sort_keys=True, ensure_ascii=False, indent=1))

        up = [
            'from migrator import load_data',
            '',
            'models = dict({})'.format(', '.join(['{}={}'.format(k, k) for k in data_to_save.keys()])),
            "load_data(config, "
            "migration_hash='{migration_hash}', "
            "models=models"
            ")".format(
                config=self.config,
                migration_hash=migration_hash,
            ),
        ]

        db_models = list(inspector.inspect_database(self.db))
        c = CodeGenerator(db_models)
        imports, models, proxies = c.prepare_code()

        return self.executor.make_migration(
            imports, models, up=up, down=None, migration_name=migration_name, proxies=proxies
        )

    def load_data(self, migration_hash, models):
        migrations_path = self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR)
        fixture_path = os.path.join(migrations_path, 'fixtures', migration_hash + '.json')
        if not os.path.exists(fixture_path):
            raise Exception('No fixture file at the {}'.format(fixture_path))
        with open(fixture_path, 'r', encoding='utf-8') as f:
            data = f.read()
        fixture = json.loads(data, encoding='utf-8', object_hook=pickle_hook)
        self._preload()
        for model_name, model_class in models.items():
            data = fixture.get(model_name, [])
            if not data:
                continue
            self._model_preload(model_class)
            for row in data:
                instance = dict_to_model(model_class, row)
                exist = model_class.select().where(instance._pk_expr()).first()
                instance.save(force_insert=not bool(exist))
            self._model_postload(model_class)
        self._postload()

    def _preload(self):
        if self.config.db_type == 'mysql':
            self.db.execute_sql('SET foreign_key_checks = 0;')

    def _postload(self):
        if self.config.db_type == 'mysql':
            self.db.execute_sql('SET foreign_key_checks = 1;')

    def _model_preload(self, model_class):
        if self.config.db_type == 'postgres':
            schema = ''
            if model_class._meta.schema:
                schema = '{}.'.format(model_class._meta.schema)
            self._with_schema = lambda x: '{}{}'.format(schema, x)
            self._table = self._with_schema(model_class._meta.table_name)
            model_class.raw(f'ALTER TABLE "{self._table}" DISABLE TRIGGER USER;').execute()

    def _model_postload(self, model_class):
        if self.config.db_type == 'postgres':
            model_class.raw(f'ALTER TABLE "{self._table}" ENABLE TRIGGER USER;').execute()
            seq = self._with_schema('{}_{}_seq'.format(model_class._meta.table_name, model_class._meta.primary_key.column_name))
            model_class.raw(
                f"SELECT setval('{seq}', (SELECT COALESCE(MAX(id)+(SELECT increment_by FROM {seq}), "
                f'(SELECT min_value FROM {seq})) FROM {self._table}), false)'
            ).execute()
        elif self.config.db_type == 'mysql':
            max_val = model_class.select(peewee.fn.Max(model_class._get_pk_value(model_class))).scalar()
            model_class.raw(f'ALTER TABLE {model_class._meta.table_name} AUTO_INCREMENT = {max_val + 1};').execute()


def load_data(config, migration_hash, models):
    loader = FixtureLoader(config)
    loader.load_data(migration_hash, models)



import datetime
import hashlib
import json
import os
import sys
import time

import peewee

from migrator.code_generator import CodeGenerator
from migrator.collector import ChangesCollector
from migrator.db_inspector import Inspector

__all__ = ['Executor']


def extend_path(path, index=0):
    for p in reversed(path.split(':')):
        if p not in sys.path:
            sys.path.insert(index, p)


database_storage = peewee.Proxy()


class DatabaseStorageBackend(peewee.Model):
    revision = peewee.CharField(max_length=32, unique=True)
    applied_at = peewee.DateTimeField(default=datetime.datetime.now)
    faked = peewee.BooleanField(default=False)

    class Meta:
        table_name = 'migrator_migrations'
        database = database_storage


class Executor(object):
    CodeGenerator = CodeGenerator
    STATUS_APPLIED = 'applied'
    STATUS_AVAILABLE = 'available'
    REQUIRED_FILE = 'required.json'

    def __init__(self, config):
        self.config = config
        # Need to correct work
        extend_path(config.get_setting(config.MIGRATOR_SYS_PATH), 1)
        self.migrations_in_path = False
        self._db_storage_initialized = False

    def init_db_storage(self):
        if self._db_storage_initialized:
            return
        database_storage.initialize(self.get_db_obj())
        DatabaseStorageBackend.create_table(safe=True)
        self._db_storage_initialized = True

    def _extend_migrations(self):
        if not self.migrations_in_path:
            extend_path(self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR), 0)
            self.migrations_in_path = True

    def import_migration(self, migration):
        return __import__(migration['import'])

    def get_migrations(self):
        migrations_dir = self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR)
        # Need to make __import__ works
        self._extend_migrations()

        for f in os.listdir(migrations_dir):
            if f.startswith('migration_') and f.endswith('.py'):
                revision = f.split('migration_', 1)[-1].rsplit('.py', 1)[0]
                yield self.fetch_migration(revision)

    def get_migrations_by_hash(self, base_hash, migrations=None):
        migrations = self.get_migrations() if migrations is None else migrations
        return [x for x in migrations if x['hash'].startswith(base_hash)]

    def fetch_migration(self, revision):
        self._extend_migrations()
        migration = __import__('migration_{}'.format(revision))
        return {
            'time': migration.MIGRATION_TIME,
            'name': migration.MIGRATION_NAME,
            'hash': revision,
            'file': 'migration_{}.py'.format(revision),
            'import': 'migration_{}'.format(revision),
            'dependencies': migration.MIGRATION_DEPENDENCIES,
            # is applied
            'status': self.check_status(revision),
            'required': self.check_required_position(revision)
        }

    def apply(self, migration, fake=False):
        if not fake:
            _module = self.import_migration(migration)
            _module.up(config=self.config)
        self.make_applied(migration['hash'], fake=fake)

    def revert(self, migration, fake=False):
        if not fake:
            _module = self.import_migration(migration)
            _module.down(self.config)
        self.unmake_applied(migration['hash'])

    def get_applied_fs(self):
        migrations_dir = self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR)
        try:
            with open(os.path.join(migrations_dir, '.applied.json'), 'r', encoding='utf-8') as f:
                return json.loads(f.read())
        except:
            return []

    def get_applied_db(self):
        self.init_db_storage()
        return [x.revision for x in DatabaseStorageBackend.select(DatabaseStorageBackend.revision)]

    def get_applied(self):
        if self.config.storage_type == self.config.STORAGE_FS:
            return self.get_applied_fs()
        else:
            return self.get_applied_db()

    def get_required(self):
        migrations_dir = self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR)
        try:
            with open(os.path.join(migrations_dir, self.REQUIRED_FILE), 'r', encoding='utf-8') as f:
                return json.loads(f.read())
        except:
            return []

    def make_required(self, revision, after=None):
        required = self.get_required()
        position = len(required)  # Last one
        if after is not None:
            try:
                position = required.index(after) + 1
            except ValueError:
                pass
        required.insert(position, revision)
        self.save_required_fs(required)

    def make_applied(self, revision, fake=False):
        if self.config.storage_type == self.config.STORAGE_FS:
            applied = set(self.get_applied())
            applied.add(revision)
            self.save_applied_fs(applied)
        else:
            self.save_applied_db(revision, fake=fake)

    def unmake_applied(self, revision):
        if self.config.storage_type == self.config.STORAGE_FS:
            applied = set(self.get_applied())
            applied.remove(revision)
            self.save_applied_fs(applied)
        else:
            self.save_applied_db(revision, revert=True)

    def save_applied_fs(self, applied):
        migrations_dir = self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR)
        with open(os.path.join(migrations_dir, '.applied.json'), 'w', encoding='utf-8') as f:
            f.write(json.dumps(list(applied)))

    def save_applied_db(self, revision, revert=False, fake=False):
        self.init_db_storage()
        if revert:
            return DatabaseStorageBackend.delete().where(DatabaseStorageBackend.revision == revision).execute() == 1
        DatabaseStorageBackend.create(revision=revision, faked=fake)
        return True

    def save_required_fs(self, required):
        migrations_dir = self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR)
        with open(os.path.join(migrations_dir, self.REQUIRED_FILE), 'w', encoding='utf-8') as f:
            f.write(json.dumps(required))

    def check_status(self, revision):
        applied = self.get_applied()

        return self.STATUS_APPLIED if revision in applied else self.STATUS_AVAILABLE

    def check_required_position(self, revision):
        required = self.get_required()

        try:
            return required.index(revision)
        except ValueError:
            return None

    def check_dependencies(self, migration):
        return all([self.check_status(dep) == self.STATUS_APPLIED for dep in migration['dependencies']])

    def get_db_obj(self):
        return self.config.get_db()

    def check_migrations_package(self):
        migrations_path = self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR)
        os.makedirs(migrations_path, exist_ok=True)
        init_path = os.path.join(migrations_path, '__init__.py')
        if not os.path.exists(init_path):
            with open(init_path, 'w') as f:
                f.write('')

    def make_empty_migration(self, migration_name=None):
        i = Inspector(models_path=self.config.get_models_paths(), excluded_models=self.config.get_excluded())
        current_models = list(i.inspect_models())
        c = CodeGenerator(current_models)
        imports, models, proxies = c.prepare_code()
        new = {x['name']: x for x in c.prepare_dict()['models']}
        return self.migrate(imports, models, proxies, new, new, migration_name=migration_name)

    def migrate_from_migration(self, migration=None, migration_name=None):
        excluded_models = self.config.get_excluded()
        # Get old state from models
        i = Inspector(models_path=[migration['import']], excluded_models=excluded_models)
        old_models = list(i.inspect_models())
        old = {x['name']: x for x in CodeGenerator(old_models).prepare_dict()['models']}
        # Get current state from models
        i = Inspector(models_path=self.config.get_models_paths(), excluded_models=excluded_models)
        current_models = list(i.inspect_models())
        new = {x['name']: x for x in CodeGenerator(current_models).prepare_dict()['models']}
        # Generate code and new models
        c = CodeGenerator(current_models + old_models)
        imports, models, proxies = c.prepare_code()
        return self.migrate(
            imports, models, proxies, new, old, migration_name=migration_name, dependencies=[migration]
        )

    def migrate_from_db(self, migration_name=None):
        excluded_models = self.config.get_excluded()
        i = Inspector(models_path=self.config.get_models_paths(), excluded_models=excluded_models)
        # Get current state from models
        current_models = list(i.inspect_models())
        new = {x['name']: x for x in CodeGenerator(current_models).prepare_dict()['models']}
        # Get state from database
        db_models = list(i.inspect_database(self.get_db_obj()))
        old = {x['name']: x for x in CodeGenerator(db_models).prepare_dict()['models']}
        # Generate code
        current_models_tables = [m[1] for m in current_models]
        for db_model in db_models:
            if db_model[1] not in current_models_tables:
                current_models.append(db_model)
        c = CodeGenerator(current_models)
        imports, models, proxies = c.prepare_code()
        return self.migrate(imports, models, proxies, new, old, migration_name=migration_name, merge_by_table=True)

    def migrate(self, imports, models, proxies, new, old, migration_name=None, dependencies=None, merge_by_table=False):
        up_migration = self._get_migration_changes(new=new, old=old)
        up = self.CodeGenerator.changes_code(up_migration, indent=4)
        down_new, down_old = old, new
        if merge_by_table:
            by_table_name = {v['table']: v['name'] for v in new.values()}
            old_model_names = {v['table']: v['name'] for v in old.values()}
            old_to_new_names = {v: by_table_name.get(k, v) for k, v in old_model_names.items()}
            down_new = {}
            for k, v in old.items():
                model_name = by_table_name.get(v['table'], v['name'])
                down_fields = []
                for f in v.get('fields', []):
                    if f['path'] != 'peewee.DeferredForeignKey':
                        down_fields.append(f)
                        continue
                    field = f.copy()
                    params = field.get('params', {}).copy()
                    arg = params['__args'][1:-1]
                    if arg in old_to_new_names:
                        arg = old_to_new_names[arg]
                    params['__args'] = f"'{arg}'"
                    field['params'] = params
                    down_fields.append(field)

                down_new[model_name] = {**v, **{'name': model_name, 'fields': down_fields}}

        down_migration = self._get_migration_changes(new=down_new, old=down_old)
        down = self.CodeGenerator.changes_code(down_migration, indent=4, revert=True)
        return self.make_migration(
            imports, models, up=up, down=down, migration_name=migration_name, proxies=proxies,
            dependencies=dependencies
        )

    def _get_migration_changes(self, new, old, revert=False):
        collector = ChangesCollector()
        model_matches = collector.get_table_matches(new, old)
        migration = {
            'drop': [x for x in old.keys() if x not in model_matches.values()],
            'create': [(x, new[x]['constraints']) for x in new.keys() if x not in model_matches.keys()],
            'rename': [],
            'fields': {
                'rename': [],
                'create': [],
                'drop': [],
                'index': [],
                'null': []
            }
        }
        for new_model_name, old_model_name in model_matches.items():
            new_model = new[new_model_name]
            old_model = old[old_model_name]
            # Table renaming case
            if new_model['table'] != old_model['table']:
                migration['rename'].append((old_model['table'], new_model['table']))

            # Fields diff
            new_fields = {x['name']: x for x in new[new_model_name]['fields']}
            old_fields = {x['name']: x for x in old[old_model_name]['fields']}

            field_matches = collector.get_field_matches(new[new_model_name]['fields'], new_fields, old_fields)
            # Fields diff ops
            migration['fields']['drop'].extend(
                collector.get_columns_to_drop(new_fields, old_fields, new_model, old_model, field_matches)
            )
            migration['fields']['create'].extend(
                collector.get_columns_to_create(new_fields, old_fields, new_model, old_model, field_matches)
            )

            for new_field_name, old_field_name in field_matches.items():
                new_field = new_fields[new_field_name]
                old_field = old_fields[old_field_name]
                migration['fields']['index'].extend(
                    collector.get_field_indexes(new_field, old_field, new_model, old_model))
                migration['fields']['null'].extend(collector.get_field_null(new_field, old_field, new_model, old_model))
                # Fields renaming case
                migration['fields']['rename'].extend(
                    collector.get_field_rename(new_field, old_field, new_model, old_model)
                )

        return migration

    def make_migration(
        self, imports, models, up=None, down=None, migration_name=None, proxies=None, dependencies=None
    ):

        self.check_migrations_package()

        migration_time = int(time.time())
        migration_hash = hashlib.md5(str(migration_time).encode('utf-8')).hexdigest()

        migration_kwargs = locals()
        migration_kwargs.pop('self', None)

        migration_path = os.path.join(
            self.config.get_setting(self.config.MIGRATOR_MIGRATIONS_DIR),
            'migration_{}.py'.format(migration_hash)
        )
        migration_code = self.CodeGenerator.migration_code(**migration_kwargs)

        with open(migration_path, 'w', encoding='utf-8') as f:
            f.write(migration_code)

        return migration_hash

    def _get_tables_as_models(self, db=None):
        db = self.get_db_obj() if db is None else db

        class BaseModel(peewee.Model):
            class Meta:
                database = db

        for table in db.get_tables():
            model = type('_', (BaseModel,), {})
            model._meta.table_name = table
            yield model

    def drop_all(self, db=None):
        for model in self._get_tables_as_models(db=db):
            model.drop_table(cascade=True)

    def truncate_all(self, db=None):
        db = self.get_db_obj() if db is None else db
        db_type = self.config.db_type
        if db_type == 'mysql':
            db.execute_sql('SET foreign_key_checks = 0;')
        for table in db.get_tables():
            if db_type in ('mysql', 'sqlite'):
                delim = '"' if db_type == 'sqlite' else '`'
                db.execute_sql(f'TRUNCATE TABLE {delim}{table}{delim}')
            else:
                db.execute_sql(f'TRUNCATE TABLE "{table}" RESTART IDENTITY CASCADE')
        if db_type == 'mysql':
            db.execute_sql('SET foreign_key_checks = 1;')



__all__ = ['CodeGenerator']


class CodeGenerator(object):
    MIGRATION_TEMPLATE = '''{imports}
from playhouse.migrate import migrate


MIGRATION_NAME = {migration_name}
MIGRATION_TIME = {migration_time}
MIGRATION_DEPENDENCIES = {migration_dependencies}


proxy_db = peewee.Proxy()


class BaseModel(peewee.Model):
    class Meta:
        database = proxy_db


{models}{proxies_init}


def up(config):
    proxy_db.initialize(config.get_db())
    migrator = config.get_migrator()
{up}


def down(config):
    proxy_db.initialize(config.get_db())
    migrator = config.get_migrator()
{down}
'''

    def __init__(self, models):
        self.models = models

    class Builder(object):
        def __init__(self, indent=0, indent_step=4):
            self.indent = indent
            self.indent_step = indent_step
            self.lines = []

        def tab(self):
            self.indent += self.indent_step
            return self

        def un_tab(self):
            self.indent -= self.indent_step
            return self

        def write_line(self, text):
            self.lines.append('{}{}'.format(' ' * self.indent, text))
            return self

        @property
        def code(self):
            return '\n'.join(self.lines)

    @classmethod
    def _join_lines_with_indent(cls, lines, indent=4):
        return '\n'.join(['{}{}'.format(' ' * indent, x) for x in lines])

    @classmethod
    def _get_import_by_path(cls, path):
        return '.'.join(path.split('.')[:-1])

    @classmethod
    def prepare_field_code(cls, field, field_name=None):
        if field_name is None:
            field_name = field['name']
        elif field['path'] == 'peewee.DeferredForeignKey':
            field['path'] = 'peewee.ForeignKeyField'
            field['params']['__args'] = field['params']['__args'].replace('"', '').replace("'", '')
            field['params']['field'] = f'{field["params"]["__args"]}.id'

        params = []
        if '__args' in field['params']:
            params.append(field['params'].pop('__args'))
        for k, v in field['params'].items():
            params.append(f'{k}={v}')
        return '{} = {}({})'.format(field_name, field['path'], ', '.join(params))

    def prepare_code(self):
        builder = self.Builder()
        to_import = {'peewee'}
        models_str = []
        prepared = self.prepare_dict()
        for model in prepared['models']:
            builder.write_line('class {}(BaseModel):'.format(model['name']))
            for field in model['fields']:
                for p in model.get('imports', []):
                    to_import.add(p)
                builder.tab().write_line(self.prepare_field_code(field.copy())).un_tab()
            builder.write_line('').tab().write_line('class Meta:')
            builder.tab().write_line('table_name = {}'.format(model['table'].__repr__())).un_tab().un_tab()
            builder.write_line('').write_line('')
        models_str.append(builder.code)
        imports_str = ['import {}'.format(x) for x in to_import]
        return imports_str, models_str, prepared['proxies']

    def prepare_dict(self):
        models_dict = []
        proxies = {}
        for model, db_table, fields in self.models:
            model_data = {'name': model, 'constraints': []}
            fields_json = []
            imports = set()
            for f in fields:
                field = f.copy()
                initial_kwargs = field['params'].get('initial_kwargs', {}).copy()
                import_path = self._get_import_by_path(field['path'])
                params = {}
                if field['params']['index'] and not field['params']['unique']:
                    params['index'] = 'True'
                if field['params']['unique']:
                    params['unique'] = 'True'
                if field['params']['null']:
                    params['null'] = 'True'
                if field['params'].get('default'):
                    field_default = field['params']['default']
                    need_import = field_default.get('import') is not None
                    params['default'] = field_default['value'] if need_import else field_default['value'].__repr__()
                    if need_import:
                        imports.add(field_default['import'])
                db_field = initial_kwargs.pop('db_field', field['column'])
                if db_field and db_field != field['name']:
                    params['column_name'] = db_field.__repr__()
                to_field = initial_kwargs.pop('to_field', {})
                if to_field and to_field['column'] != 'id':
                    params['field'] = to_field['column'].__repr__()
                if initial_kwargs:
                    rel = initial_kwargs.pop('rel_model', None)
                    if rel:
                        if rel == model:
                            rel = 'self'
                        else:
                            proxies.setdefault(rel, set()).add(f'{model}.{field["name"]}')
                            model_data['constraints'].append(field['name'])
                            field['path'] = 'peewee.DeferredForeignKey'
                        params['__args'] = rel.__repr__()

                    for k, v in initial_kwargs.items():
                        if not isinstance(v, dict):
                            params[k] = v.__repr__()
                            continue
                        need_import = v.get('import') is not None
                        params[k] = v['value'] if need_import else v['value'].__repr__()
                        if need_import:
                            imports.add(v['import'])

                field_json = {'path': field['path'], 'params': params, 'name': field['name']}
                if field['fk_name'] is not None:
                    field_json['fk_name'] = field['fk_name']
                if import_path != 'peewee':
                    field_json['import'] = import_path
                    imports.add(import_path)
                fields_json.append(field_json)
            model_data.update({'table': db_table, 'fields': fields_json})
            if imports:
                model_data['imports'] = list(imports)
            models_dict.append(model_data)
        return {
            'proxies': {k: list(v) for k, v in proxies.items()},
            'models': sorted(models_dict, key=lambda x: x['name'])
        }

    @classmethod
    def changes_code(cls, migration, indent=0, revert=False):
        builder = cls.Builder(indent=indent)
        new_constraints = []

        if migration['create']:
            builder.write_line('').write_line('# Tables creation')
            for model, constraints in migration['create']:
                builder.write_line(f'{model}.create_table()')
                new_constraints.extend((model, c) for c in constraints)
        if new_constraints:
            builder.write_line('').write_line('# New constraints creation')
            builder.write_line("if config.db_type != 'sqlite':").tab()
            for model, c in new_constraints:
                builder.write_line(f'{model}._schema.create_foreign_key({model}.{c})')
            builder.un_tab()

        if migration['drop']:
            builder.write_line('').write_line('# Tables deletion')
            for db_table in migration['drop']:
                builder.write_line(f'{db_table}.drop_table()')

        # Tables renaming
        pass

        if migration['fields']['create']:
            builder.write_line('').write_line('# Fields creation')
            if revert:
                # this fields not in migration code
                for i, (table, db_field, field) in enumerate(migration['fields']['create'], 1):
                    new_field = f'DeletedField{i}'
                    builder.write_line(cls.prepare_field_code(field["field"], field_name=new_field))
                    field['field_string'] = new_field
            builder.write_line('migrate(').tab()
            for table, db_field, field in migration['fields']['create']:
                builder.write_line(
                    'migrator.add_column({}, {}, {}),'.format(table.__repr__(), db_field, field['field_string'])
                )

            builder.un_tab().write_line(')')

        if migration['fields']['drop']:
            builder.write_line('').write_line('# Drop fields')
            builder.write_line('migrate(').tab()
            for table, db_field in migration['fields']['drop']:
                builder.write_line('migrator.drop_column({}, {}),'.format(table.__repr__(), db_field))
            builder.un_tab().write_line(')')

        # Fields renaming
        pass

        to_drop = [x for x in migration['fields']['index'] if x[0] == 'drop']
        if to_drop:
            builder.write_line('').write_line('# Drop indexes')
            builder.write_line('migrate(').tab()
            for action, table, db_field in to_drop:
                builder.write_line('migrator.drop_index({}, {}),'.format(table.__repr__(), db_field.__repr__()))
            builder.un_tab().write_line(')')

        to_create = [x for x in migration['fields']['index'] if x[0] == 'add']
        unique = [x for x in to_create if x[-1]]
        index = [x for x in to_create if not x[-1]]
        if index:
            builder.write_line('').write_line('# Create indexes')
            builder.write_line('migrate(').tab()
            for action, table, db_field, is_uniq in index:
                builder.write_line(
                    'migrator.add_index({}, ({},), False),'.format(table.__repr__(), db_field.__repr__())
                )
            builder.un_tab().write_line(')')
        if unique:
            builder.write_line('').write_line('# Create unique indexes')
            builder.write_line('migrate(').tab()
            for action, table, db_field, is_uniq in index:
                builder.write_line(
                    'migrator.add_index({}, ({},), True)),'.format(table.__repr__(), db_field.__repr__())
                )
            builder.un_tab().write_line(')')

        result_code = [builder.code]
        return result_code

    @classmethod
    def migration_code(
        cls, imports, models, up=None, down=None, migration_name=None, proxies=None, dependencies=None,
        migration_time=None, migration_hash=None
    ):
        if migration_name is None:
            migration_name = 'Migration {}'.format(migration_hash)
        if dependencies is None:
            migration_dependencies = []
        else:
            migration_dependencies = '[{}]'.format(', '.join([x['hash'].__repr__() for x in dependencies]))
        if up is None:
            up = ['pass']
        if down is None:
            down = ['pass']
        up = cls._join_lines_with_indent(up)
        down = cls._join_lines_with_indent(down)

        proxy_fields_init = []
        for proxy, fields in proxies.items():
            proxy_fields_init.append(f'peewee.DeferredForeignKey.resolve({proxy})')
            # for f in fields:
            #     proxy_fields_init.append(f'{f}.set_model({proxy})')
        proxy_fields_init = '\n{}'.format('\n'.join(proxy_fields_init)) if proxy_fields_init else ''

        return cls.MIGRATION_TEMPLATE.format(
            up=up, down=down, imports='\n'.join(imports), models='\n\n\n'.join(models),
            migration_name=migration_name.__repr__(), migration_dependencies=migration_dependencies,
            migration_time=migration_time, proxies_init=proxy_fields_init
        )



import inspect

from inspect import getmodule, ismethod, isclass

import peewee
from peewee import ModelBase, ForeignKeyField
from playhouse.reflection import Introspector, UnknownField

from importlib import reload

__all__ = ['Inspector', 'DEFAULT_MIGRATOR_PARAMS']


DEFAULT_MIGRATOR_PARAMS = {
    peewee.CharField: (
        ('max_length', 255),
    )
}


class CustomReflection(Introspector):
    def generate_models(self, skip_invalid=False, table_names=None,
                        literal_column_names=False, bare_fields=False):
        database = self.introspect(table_names=table_names,
                                   literal_column_names=literal_column_names)
        models = {}

        class BaseModel(peewee.Model):
            class Meta:
                database = self.metadata.database
                schema = self.schema

        reflected = set()  # migrator changes

        def _create_model(table, models):
            for foreign_key in database.foreign_keys[table]:
                dest = foreign_key.dest_table

                # migrator changes
                if dest not in models and dest != table and dest not in reflected:
                    # Stop infinite recursion
                    reflected.add(dest)
                    # / migrator changes
                    _create_model(dest, models)

            primary_keys = []
            columns = database.columns[table]
            for column_name, column in columns.items():
                if column.primary_key:
                    primary_keys.append(column.name)

            multi_column_indexes = database.multi_column_indexes(table)
            column_indexes = database.column_indexes(table)

            class Meta:
                indexes = multi_column_indexes
                table_name = table

            # Fix models with multi-column primary keys.
            composite_key = False
            if len(primary_keys) == 0:
                primary_keys = columns.keys()
            if len(primary_keys) > 1:
                Meta.primary_key = peewee.CompositeKey(*[
                    field.name for col, field in columns.items()
                    if col in primary_keys])
                composite_key = True

            attrs = {'Meta': Meta}
            for column_name, column in columns.items():
                FieldClass = column.field_class
                if FieldClass is not ForeignKeyField and bare_fields:
                    FieldClass = peewee.BareField
                elif FieldClass is UnknownField:
                    FieldClass = peewee.BareField

                params = {
                    'column_name': column_name,
                    'null': column.nullable}
                if column.primary_key and composite_key:
                    if FieldClass is peewee.AutoField:
                        FieldClass = peewee.IntegerField
                    params['primary_key'] = False
                elif column.primary_key and FieldClass is not peewee.AutoField:
                    params['primary_key'] = True
                if column.is_foreign_key():
                    if column.is_self_referential_fk():
                        params['model'] = 'self'
                    else:
                        dest_table = column.foreign_key.dest_table
                        # migrator changes
                        if dest_table not in models:
                            params['rel_model_name'] = database.model_names[dest_table]
                            FieldClass = peewee.DeferredForeignKey
                        else:
                            params['model'] = models[dest_table]
                        # / migrator changes
                    if column.to_field:
                        params['field'] = column.to_field

                    # Generate a unique related name.
                    params['backref'] = '%s_%s_rel' % (table, column_name)

                if column.default is not None:
                    constraint = peewee.SQL('DEFAULT %s' % column.default)
                    params['constraints'] = [constraint]

                if column_name in column_indexes and not \
                   column.is_primary_key():
                    if column_indexes[column_name]:
                        params['unique'] = True
                    elif not column.is_foreign_key():
                        params['index'] = True

                attrs[column.name] = FieldClass(**params)

            try:
                models[table] = type(str(database.model_names[table]), (BaseModel,), attrs)  # migrator changes
            except ValueError:
                if not skip_invalid:
                    raise

        # Actually generate Model classes.
        for table, model in sorted(database.model_names.items()):
            if table not in models:
                _create_model(table, models)

        return models


class Inspector(object):
    DEFAULT_PARAMS = (
        ('sequence', None),
        ('default', None),
        ('constraints', None),
        ('schema', None),
        ('on_delete', None),
        ('on_update', None),
        ('extra', None),
        ('primary_key', False)
    )
    models_path = ('app.models',)

    def __init__(self, models_path=None, excluded_models=None):
        if models_path is not None:
            self.models_path = models_path
        self.excluded_models = [] if excluded_models is None else excluded_models
        self._collected_models = set()

    @staticmethod
    def is_model(obj):
        return isclass(obj) and (issubclass(obj, ModelBase) or isinstance(obj, ModelBase))

    def get_models(self):
        for path in self.models_path:
            module = __import__(path, fromlist=['*'])
            models = reload(module)
            for model in dir(models):
                if model[0].isupper() and model[0] != '_':
                    obj = getattr(models, model)
                    if self.is_model(obj) and model not in self.excluded_models:
                        yield model, obj

    def get_database_models(self, db=None):
        i = CustomReflection.from_database(db)
        return i.generate_models().values()

    def inspect_models(self):
        for model, model_obj in self.get_models():
            if model in self._collected_models:
                # prevent doubles for imported models
                continue
            self._collected_models.add(model)
            yield self.inspect_model(model_obj, model=model)
        self._collected_models = set()

    def inspect_database(self, db):
        models = self.get_database_models(db)
        for model_obj in models:
            if model_obj.__name__ in self.excluded_models:
                continue
            yield self.inspect_model(model_obj)

    def inspect_model(self, model_obj, model=None):
        if model is None:
            model = model_obj.__name__
        fields = list(self.collect_fields(model_obj._meta.sorted_fields, model=model))
        return model, model_obj._meta.table_name, fields

    @classmethod
    def get_param_value(cls, value):
        primitive_types = (int, str, bytes, bool)
        if isinstance(value, primitive_types):
            # Raw type
            return {'value': value}

        if not callable(value):
            # Mutable type, do not apply automatically
            return None
        _module = inspect.getmodule(value)
        if _module is not None:
            # Simple functions
            module_name = _module.__name__
            return {'value': '{}.{}'.format(module_name, value.__name__), 'import': module_name}
        # datetime case
        if value.__name__ == 'now' and type(value).__name__ == 'builtin_function_or_method':
            return {'value': 'datetime.datetime.now', 'import': 'datetime'}
        # TODO: Process builtins with _module = None
        return None

    @classmethod
    def get_field_initial_params(cls, field):
        param_keys = [*cls.DEFAULT_PARAMS]
        param_keys.extend(getattr(field, 'MIGRATOR_PARAMS', ()))
        if type(field) in DEFAULT_MIGRATOR_PARAMS:
            param_keys.extend(DEFAULT_MIGRATOR_PARAMS[type(field)])
        return tuple(set(param_keys))

    def collect_fields(self, sorted_fields, model=''):
        for field in sorted_fields:
            is_fk = isinstance(field, ForeignKeyField)
            name = field.name
            if name == 'id' and isinstance(field, peewee.AutoField):
                # Skip auto-generated id
                continue
            field_path = field.__class__.__name__
            field_module = getmodule(field.__class__)
            if field_module is not None:
                field_module = field_module.__name__
                field_path = '{}.{}'.format(field_module, field_path)

            default_value = self.get_param_value(field.default)
            field_params = {
                'index': field.index, 'unique': field.unique, 'null': field.null,
                'initial_kwargs': {}
            }
            if default_value is not None:
                field_params['default'] = default_value

            for param in self.get_field_initial_params(field):
                if not isinstance(param, str):
                    param, default_val = param
                    param_obj = getattr(field, param, None)
                    if (default_val is None and param_obj is None) or (default_val is False and param_obj is False):
                        continue
                    if param_obj == default_val:
                        continue
                else:
                    param_obj = getattr(field, param, None)
                if ismethod(param_obj):
                    continue

                field_params['initial_kwargs'][param] = self.get_param_value(param_obj)

            if field.name != field.column_name:
                if not is_fk or (is_fk and field.column_name != f'{field.name}_id'):
                    field_params['initial_kwargs']['column_name'] = field.column_name

            db_column = field.column_name

            if isinstance(field, ForeignKeyField):
                # Save only model name
                field_params['initial_kwargs']['rel_model'] = field.rel_field.model.__name__
                # Field normalization
                db_column = field.name if field.column_name == f'{field.name}_id' else field.column_name
                to_field = {
                    'name': field.rel_field.name, 'column': field.rel_field.column_name
                }
                if to_field['name'] == 'id' and to_field['column'] == 'id':
                    field_params['initial_kwargs'].pop('to_field', None)
                else:
                    field_params['initial_kwargs']['to_field'] = to_field

            if not field_params['initial_kwargs']:
                field_params.pop('initial_kwargs', None)

            yield {
                'name': name, 'column': db_column, 'path': field_path, 'params': field_params,
                'fk_name': field.column_name if is_fk else None
            }



import base64
import json
import pickle


class SafeEncoder(json.JSONEncoder):
    def default(self, obj):
        try:
            return super(SafeEncoder, self).default(self, obj)
        except TypeError:
            try:
                if isinstance(obj, memoryview):
                    obj = bytes(obj)
                return {'__pickled_object__': base64.b64encode(pickle.dumps(obj)).decode()}
            except:
                return '__unencoded_object__'


def pickle_hook(dct):
    if '__pickled_object__' in dct:
        try:
            return pickle.loads(base64.b64decode(dct['__pickled_object__']))
        except:
            return '__unencoded_object__'
    return dct
